%
%   This function solves the degree constrained Nevanlinna-Pick 
%    interpolation problem for a desired set of spectral zeros.
%    Let m be the number of interpolation points taken with multi-
%    plicity minus one.
%
%   Input: interdata - interpoltion data set
%
%   Output: iterpolant - a interpolant data set
%
function interpolant = npdc(interdata)

%
%   Parameters
%
error_bound = 1e-5; %   maximal 2-norm of interpolation errors 
epsilon = 5e-2;     %   a number which determines predictor-step size
                    %   (ellipsoidal size)
slope = 3e-1;       %   slope bound in Wolfe test
                    %   large --> more inaccurate line search
digit = 8;          %   for display

%
%   Make data self-conjugated, move first interpolation point to zero
%    and check that data is valid.
%
[interdata, tozerotrans] = datavalid(interdata);

%
%   Determine Pick matrix
%
P = getpickmatrix(interdata);
if min(eig(P))<=0,
  disp('Pick matrix is NOT positive!')
  return
end

%
%   Generate L, V & K matrices
%
L = toeplitz(interdata.tau,[1 ,zeros(1,interdata.m-1)]);
V = getV(interdata);
invK = real(L*V*inv(P.')*V'*L');   % L is real so L' = L.'
K = inv(invK);

%
%   Determine central solution
%
alpha0 = sqrt(invK(1,1));
alpha = invK(2:end,1) / alpha0;
alpha = [ alpha0 alpha' ];
beta = getbeta(alpha,L'*interdata.tau');
error_int = interror(alpha,beta,interdata);

%
%   Determine the rho polynomial
%
if isempty(interdata.speczeros)
    rho = interdata.tau;
else
    rho = fliplr(poly(interdata.speczeros));
    rho = [ zeros(1,length(interdata.tau)-length(rho)) rho ];
end

%
%   Check whether we are looking for the central solution
%
if norm(rho-interdata.tau) < eps
   disp('This is the central solution.')
   error_int = interror(alpha,beta,interdata);
   if error_int > error_bound,
       disp('Solution might be inaccurate'), error_int
   end
   interpolant = getinterpolant(beta,alpha);
   interpolant = bilineardomain(tozerotrans,interpolant);
   return
end

%
%   Determine the phi at the desired solution and normalization
%
psi = covar2(rho,interdata.tau,interdata.m-1);
phi = psi/psi(1);

%
%   Hessian of g_rho at the central solution
%
muvec = psi2mu([phi(1) 0*phi(2:end)],interdata.tau);
h_f = hessian3(alpha,muvec,interdata.tau,K);
if min(eig(h_f)) <= 0,
  disp('Hessian is not positive definite!')
  disp('Central solution might be inaccurate.')
  error_int = interror(alpha,beta,interdata)
end

%
%   Calculation of eta * tau* + eta* * tau = muvec1
%
eta = [ 0; L(1:end-1,1:end-1)*phi(2:end)' ];
muvec1 = (hankel(interdata.tau) + L') * eta;

disp(' ')
disp('------------------------------------------------------')
disp('Nevanlinna-Pick solver based on a continuation method ')
disp('------------------------------------------------------')
disp('                       Corrector step                 ')
disp('     nu               iteration   error               ')
disp('------------------------------------------------------')

nu = 0;
while nu ~= 1

    %
    %   Predictor step:
    %    Calculate the direction for a step in nu
    %
    d_nu_grad_g = grad3(alpha,muvec1,interdata.tau,K) - 2*K*alpha(:);
    d_a_nu = - h_f \ d_nu_grad_g;

    %
    %   Determine the step size for nu
    %
    if norm(d_a_nu) == 0
        dnu = eps;
    else
        dnu = sqrt(epsilon/(d_a_nu'*K*d_a_nu));
    end
    
    if nu + dnu > 1,
        dnu = 1 - nu;
    end

    muvec = psi2mu([phi(1) (nu+dnu)*phi(2:end)],interdata.tau);
 
    %
    %   We implicitly have to fulfill
    %       1. alpha stable in the unit disc
    %       2. convex region, ie positive definite hessian
    %
    while min(abs(roots(fliplr(alpha + d_a_nu'*dnu)))) <= 1 | ...
            min(eig(hessian3(alpha + d_a_nu'*dnu,muvec,interdata.tau,K))) <= 0
        dnu = 0.5*dnu;
        muvec = psi2mu([phi(1) (nu+dnu)*phi(2:end)],interdata.tau);
    end
    nu = nu + dnu;
    disp(num2str(nu,digit))

    %
    %   Corrector step:
    %
    %       Update the polynomial alpha, generate beta and 
    %        check interpolation error.
    %
    alpha = alpha + d_a_nu'*dnu;
    beta = getbeta(alpha,muvec);
    error_int = interror(alpha,beta,interdata);

    iter = 0; 
    error_int_old = 0;
    h_f = hessian3(alpha,muvec,interdata.tau,K);
    g_f = grad3(alpha,muvec,interdata.tau,K);
    while error_int > error_bound %& norm(g_f) > 1e-8

        d_a = -(h_f\g_f)';
 
        %
        %   One-sided Wolfe test. We also check that we
        %       1. stay in the region of stable alpha
        %       2. stay in the conevx region (pos def hessian)
        %   We try to take a full Newton step and use interval halfing
        %    until we fulfill the criteria above.
        %
        lambda = 1;
        g_f_new = grad3(alpha+lambda*d_a,muvec,interdata.tau,K);
        h_f_new = hessian3(alpha+lambda*d_a,muvec,interdata.tau,K);
        
        pprim0 = d_a * g_f;
        pprimnew = d_a * g_f_new;
        pbisnew = d_a * h_f_new * d_a';
        
        while -slope*pprim0 < pprimnew | ...
            min(abs(roots(fliplr(alpha + lambda*d_a)))) <= 1 | ...
            min(eig(h_f_new)) <= 0

%            disp([-slope*pprim0 < pprimnew min(abs(roots(fliplr(alpha + lambda*d_a)))) <= 1  min(eig(h_f_new)) <= 0 ])
             
            lambda = lambda / 2;
            g_f_new = grad3(alpha + lambda*d_a,muvec,interdata.tau,K);
            h_f_new = hessian3(alpha + lambda*d_a,muvec,interdata.tau,K);
            pprimnew = d_a * g_f_new;
        
        end

        %
        %   Update the solution and check interpolation error
        %
        alpha = alpha + lambda*d_a;
        h_f = h_f_new;
        g_f = g_f_new;
        beta = getbeta(alpha,muvec);
        error_int = interror(alpha,beta,interdata);

        iter = iter + 1;
        if mod(iter,10) == 0
                                        % Modified by Anders 020116
            if abs(error_int_old - error_int) < 1e-10
                if nu == 1,
                   disp('           Solution may be inaccurate.'), break
                else
                   disp('  No improvement. Next predictor step.'), break
                end
            end
            error_int_old = error_int; 

            iter_str = num2str(iter);
            error_int_str = num2str(error_int,digit);
            disp([blanks(26 - length(iter_str)) iter_str,blanks(6) ...
                    ,error_int_str])
        end
    end
end

disp('-------------------------------------------------')
disp(' ')

interpolant = getinterpolant(beta,alpha);
interpolant = bilineardomain(tozerotrans,interpolant);

disp('Finished')